package Thruk::BP::Components::BP;

use strict;
use warnings;
use Data::Dumper;
use Carp;
use File::Temp;
use File::Copy qw/move/;
use Fcntl qw/:DEFAULT/;
use Scalar::Util qw/weaken isweak/;
use Thruk::Utils;
use Thruk::Utils::IO;
use Thruk::BP::Components::Node;
use Time::HiRes qw/gettimeofday tv_interval/;

=head1 NAME

Thruk::BP::Components::Node - BP Class

=head1 DESCRIPTION

Business Process

=head1 METHODS

=cut

my @extra_json_keys = qw/id/;
my @stateful_keys   = qw/status status_text last_check last_state_change time/;
my @saved_keys      = qw/name template rankDir state_type/;

##########################################################

=head2 new

return new business process

=cut

sub new {
    my ( $class, $c, $file, $bpdata, $editmode ) = @_;

    my $self = {
        'id'                => undef,
        'editmode'          => $editmode,
        'name'              => undef,
        'template'          => $bpdata->{'template'},
        'nodes'             => [],
        'nodes_by_id'       => {},
        'nodes_by_name'     => {},
        'need_update'       => {},
        'need_save'         => 0,
        'file'              => undef,
        'datafile'          => undef,
        'editfile'          => undef,

        'time'              => 0,
        'status'            => 4,
        'status_text'       => 'not yet checked',
        'last_check'        => 0,
        'last_state_change' => 0,
        'rankDir'           => 'TB',
        'state_type'        => 'both',

        'exported_nodes'    => {},
        'testmode'          => 0,
    };
    bless $self, $class;
    $self->set_file($c, $file);

    if($editmode and -e $self->{'editfile'}) { $file = $self->{'editfile'}; }
    if(-e $file) {
        $bpdata = Thruk::Utils::IO::json_lock_retrieve($file);
        return unless $bpdata;
        return unless $bpdata->{'name'};
    }
    for my $key (@saved_keys) {
        $self->{$key} = $bpdata->{$key} if defined $bpdata->{$key};
    }
    $self->set_label($c, $bpdata->{'name'});


    return unless $self->{'name'};

    # read in nodes
    for my $n (@{Thruk::Utils::list($bpdata->{'nodes'} || [])}) {
        my $node = Thruk::BP::Components::Node->new($n);
        $self->add_node($node, 1);
    }

    $self->_resolve_nodes();

    if($self->{'editmode'}) {
        die("no context") unless $c;
        $self->update_status($c);
    } else {
        $self->load_runtime_data();
    }

    $self->save() if $self->{'need_save'};

    # avoid circular refs
    for my $n (@{$self->{'nodes'}}) {
        weaken($self->{'nodes_by_id'}->{$n->{'id'}})      unless isweak($self->{'nodes_by_id'}->{$n->{'id'}});
        weaken($self->{'nodes_by_name'}->{$n->{'label'}}) unless isweak($self->{'nodes_by_name'}->{$n->{'label'}});
    }

    return $self;
}

##########################################################

=head2 load_runtime_data

update runtime data

=cut
sub load_runtime_data {
    my($self) = @_;

    return unless -e $self->{'datafile'};

    my $data = Thruk::Utils::IO::json_lock_retrieve($self->{'datafile'});

    for my $key (@stateful_keys) {
        $self->{$key} = $data->{$key} if defined $data->{$key};
    }

    for my $n (@{$self->{'nodes'}}) {
        $n->load_runtime_data($data->{'nodes'}->{$n->{'id'}});
    }
    return;
}

##########################################################

=head2 update_status

    update_status($c, [$type])

type:
    0 / undef:      update everything
    1:              only recalculate

update status of business process

=cut
sub update_status {
    my ( $self, $c, $type ) = @_;
    die("no context") unless $c;

    my $t0 = [gettimeofday];

    $type = 0 unless defined $type;
    my $last_state = $self->{'status'};

    my $results = [];
    my($livedata);
    if($type == 0) {
        $livedata = $self->bulk_fetch_live_data($c);
        for my $n (@{$self->{'nodes'}}) {
            my $r = $n->update_status($c, $self, $livedata);
            push @{$results}, $r if $r;
        }
    }

    my $iterations = 0;
    while(scalar keys %{$self->{'need_update'}} > 0) {
        $iterations++;
        for my $id (keys %{$self->{'need_update'}}) {
            my $r = $self->{'nodes_by_id'}->{$id}->update_status($c, $self, $livedata, $type);
            push @{$results}, $r if $r;
        }
        die("circular dependenies? Still have these on the update list: ".Dumper($self->{'need_update'})) if $iterations > 10;
    }

    # update last check time
    my $now = time();
    $self->{'last_check'} = $now;
    if($last_state != $self->{'status'}) {
        $self->{'last_state_change'} = $now;
    }

    # everything else is non-edit only
    return if $self->{'editmode'} or $self->{'testmode'};

    # submit back to core
    $self->_submit_results_to_core($c, $results);

    # save runtime
    $self->{'time'} = tv_interval($t0);

    # store runtime data
    $self->save_runtime();

    return;
}

##########################################################

=head2 set_label

set label for this business process

=cut
sub set_label {
    my($self, $c, $label) = @_;
    $self->{'name'} = $label;
    return;
}

##########################################################

=head2 set_status

set status for this business process

=cut
sub set_status {
    my($self, $state, $text) = @_;

    my $last_state = $self->{'status'};

    # update last check time
    my $now = time();
    $self->{'last_check'} = $now;

    $self->{'status'}      = $state;
    $self->{'status_text'} = $text;

    if($last_state != $state) {
        $self->{'last_state_change'} = $now;
    }
    return;
}

##########################################################

=head2 set_file

set file for this business process

=cut
sub set_file {
    my($self, $c, $file) = @_;
    my $basename = $file;
    $basename    =~ s/^.*\///mx;
    $self->{'file'}     = Thruk::BP::Utils::base_folder($c).'/'.$basename;
    $self->{'datafile'} = $c->config->{'var_path'}.'/bp/'.$basename.'.runtime';
    $self->{'editfile'} = $c->config->{'var_path'}.'/bp/'.$basename.'.edit';
    $basename =~ m/(\d+).tbp/mx;
    $self->{'id'}       = $1;
    return;
}

##########################################################

=head2 get_node

return node by id

=cut
sub get_node {
    my ( $self, $node_id ) = @_;
    return $self->{'nodes_by_id'}->{$node_id};
}

##########################################################

=head2 add_node

add new node to business process

=cut
sub add_node {
    my ( $self, $node, $init ) = @_;
    push @{$self->{'nodes'}}, $node;
    if(!$init) {
        # verify uniq id
        my $id = $node->{'id'} || $self->make_new_node_id();
        if($self->{'nodes_by_id'}->{$id}) {
            $node->set_id($self->make_new_node_id());
        }
        $node->set_id($id);
    }
    $self->{'nodes_by_id'}->{$node->{'id'}}      = $node if $node->{'id'};
    $self->{'nodes_by_name'}->{$node->{'label'}} = $node;

    if($self->{'exported_nodes'}->{$node->{'label'}}) {
        $node->{'create_obj_ok'} = 0;
        $node->{'create_obj'}    = 0;
    } elsif($node->{'create_obj'}) {
        $self->{'exported_nodes'}->{$node->{'label'}} = 1;
    }

    return;
}

##########################################################

=head2 remove_node

remove node from business process

=cut
sub remove_node {
    my ( $self, $node_id ) = @_;
    my $node = $self->{'nodes_by_id'}->{$node_id};
    delete $self->{'nodes_by_id'}->{$node_id};
    delete $self->{'nodes_by_name'}->{$node->{'label'}};

    # remove connections
    for my $p (@{$node->{'parents'}}) {
        my @depends;
        for my $d (@{$p->{'depends'}}) {
            push @depends, $d unless $d->{'id'} eq $node_id;
        }
        $p->{'depends'} = \@depends;
    }

    for my $d (@{$node->{'depends'}}) {
        my @parents;
        for my $p (@{$d->{'parents'}}) {
            push @parents, $p unless $p->{'id'} eq $node_id;
        }
        $d->{'parents'} = \@parents;
    }

    my @nodes;
    for my $n (@{$self->{'nodes'}}) {
        push @nodes, $n unless $n->{'id'} eq $node_id;
    }
    $self->{'nodes'} = \@nodes;

    if($node->{'create_obj'}) {
        delete $self->{'exported_nodes'}->{$node->{'label'}};
    }

    return;
}

##########################################################
# replace list of node names/ids with references
sub _resolve_nodes {
    my($self) = @_;

    for my $node (@{$self->{'nodes'}}) {
        # make sure we have an id now
        if(!$node->{'id'}) {
            $node->set_id($self->make_new_node_id());
            $self->{'nodes_by_id'}->{$node->{'id'}} = $node;
        }
        $node->resolve_depends($self);
    }

    return;
}

##########################################################

=head2 remove

remove business process data to file

=cut
sub remove {
    my ( $self ) = @_;
    unlink($self->{'file'});     # may not exist, if removed before first commit
    unlink($self->{'datafile'}); # can fail if not updated before removal
    unlink($self->{'editfile'}); # may also not exist
    return;
}

##########################################################

=head2 commit

commit business process data to file

=cut
sub commit {
    my ( $self, $c ) = @_;

    # run pre hook
    if($c->config->{'Thruk::Plugin::BP'}->{'pre_save_cmd'}) {
        local $ENV{REMOTE_USER} = $c->stash->{'remote_user'};
        local $SIG{CHLD}        = 'DEFAULT';
        system($c->config->{'Thruk::Plugin::BP'}->{'pre_save_cmd'}, 'pre', $self->{'file'});
        if($? == -1) {
            Thruk::Utils::set_message( $c, 'fail_message', 'pre save hook failed: '.$?.': '.$! );
            return;
        }
    }

    if(-e $self->{'editfile'}) {
        move($self->{'editfile'}, $self->{'file'}) or die('cannot commit changes to '.$self->{'file'}.': '.$!);
        unlink($self->{'editfile'});
    }

    # run post hook
    if($c->config->{'Thruk::Plugin::BP'}->{'post_save_cmd'}) {
        local $ENV{REMOTE_USER} = $c->stash->{'remote_user'};
        local $SIG{CHLD}        = 'DEFAULT';
        system($c->config->{'Thruk::Plugin::BP'}->{'post_save_cmd'}, 'post', $self->{'file'});
        if($? == -1) {
            Thruk::Utils::set_message( $c, 'fail_message', 'post save hook failed: '.$?.': '.$! );
            return;
        }
    }

    return 1;
}

##########################################################

=head2 save

save business process data to temporary edit file

=cut
sub save {
    my ( $self, $c ) = @_;

    return if $self->{'testmode'};

    my $obj = {
        nodes     => [],
    };

    for my $key (@saved_keys) {
        $obj->{$key} = $self->{$key};
    }

    for my $n (@{$self->{'nodes'}}) {
        push @{$obj->{'nodes'}}, $n->get_save_obj();
    }

    Thruk::Utils::IO::json_lock_store($self->{'editfile'}, $obj, 1);
    $self->{'need_save'} = 0;

    return 1;
}

##########################################################

=head2 get_objects_conf

return object config

=cut
sub get_objects_conf {
    my ( $self ) = @_;
    my $obj = {
        'hosts'    => {},
        'services' => {},
    };
    for my $n (@{$self->{'nodes'}}) {
        my $nodedata = $n->get_objects_conf($self);
        Thruk::BP::Utils::merge_obj_hash($obj, $nodedata) if $nodedata;
    }
    return $obj;
}

##########################################################

=head2 save_runtime

save run time data

=cut
sub save_runtime {
    my ( $self ) = @_;
    return if $self->{'testmode'};
    return if $self->{'editmode'};
    my $data = {};
    for my $key (@stateful_keys) {
        $data->{$key} = $self->{$key};
    }
    for my $n (@{$self->{'nodes'}}) {
        $data->{'nodes'}->{$n->{'id'}} = $n->get_stateful_data();
    }
    Thruk::Utils::IO::json_lock_store($self->{'datafile'}, $data);
    return;
}

##########################################################

=head2 make_new_node_id

generate new uniq id

=cut
sub make_new_node_id {
    my ( $self ) = @_;
    my $num = 1;
    my $id  = 'node'.$num;
    while($self->{'nodes_by_id'}->{$id}) {
        $id = 'node'.++$num;
    }
    $self->{'need_save'} = 1;
    return $id;
}

##########################################################

=head2 bulk_fetch_live_data

return all live data needed for this business process

=cut
sub bulk_fetch_live_data {
    my($self, $c) = @_;

    # bulk fetch live data
    my $hostfilter         = {};
    my $servicefilter      = {};
    my $hostgroupfilter    = {};
    my $servicegroupfilter = {};
    my $hostdata           = {};
    my $servicedata        = {};
    my $hostgroupdata      = {};
    my $servicegroupdata   = {};
    for my $n (@{$self->{'nodes'}}) {
        if(lc $n->{'function'} eq 'status') {
            if($n->{'host'} and $n->{'service'}) {
                $servicefilter->{$n->{'host'}}->{$n->{'service'}} = 1;
            }
            elsif($n->{'host'}) {
                $hostfilter->{$n->{'host'}} = 1;
            }
        }
        elsif(lc $n->{'function'} eq 'groupstatus') {
            if($n->{'hostgroup'}) {
                $hostgroupfilter->{$n->{'hostgroup'}} = 1;
            }
            elsif($n->{'servicegroup'}) {
                $servicegroupfilter->{$n->{'servicegroup'}} = 1;
            }
        }
    }
    if(scalar keys %{$hostfilter} > 0) {
        my @filter;
        for my $hostname (keys %{$hostfilter}) {
            push @filter, { name => $hostname };
        }
        my $filter = Thruk::Utils::combine_filter( '-or', \@filter );
        my $data   = $c->{'db'}->get_hosts(filter => [$filter], extra_columns => [qw/last_hard_state last_hard_state_change/]);
        $hostdata  = Thruk::Utils::array2hash($data, 'name');
    }
    if(scalar keys %{$servicefilter} > 0) {
        my @filter;
        for my $hostname (keys %{$servicefilter}) {
            for my $description (keys %{$servicefilter->{$hostname}}) {
                push @filter, { host_name => $hostname, description => $description };
            }
        }
        my $filter = Thruk::Utils::combine_filter( '-or', \@filter );
        my $data   = $c->{'db'}->get_services(filter => [$filter], extra_columns => [qw/last_hard_state last_hard_state_change/]);
        $servicedata = Thruk::Utils::array2hash($data, 'host_name', 'description');
    }
    if(scalar keys %{$hostgroupfilter} > 0) {
        my @filter;
        for my $hostgroupname (keys %{$hostgroupfilter}) {
            push @filter, { name => $hostgroupname };
        }
        my $filter = Thruk::Utils::combine_filter( '-or', \@filter );
        my $data   = $c->{'db'}->get_hostgroups(filter => [$filter], columns => [qw/name num_hosts num_hosts_down num_hosts_pending
                                                                                 num_hosts_unreach num_hosts_up num_services num_services_crit
                                                                                 num_services_ok num_services_pending num_services_unknown
                                                                                 num_services_warn worst_service_state worst_host_state/]);
        $hostgroupdata  = Thruk::Utils::array2hash($data, 'name');
    }
    if(scalar keys %{$servicegroupfilter} > 0) {
        my @filter;
        for my $servicegroupname (keys %{$servicegroupfilter}) {
            push @filter, { name => $servicegroupname };
        }
        my $filter = Thruk::Utils::combine_filter( '-or', \@filter );
        my $data   = $c->{'db'}->get_servicegroups(filter => [$filter], columns => [qw/name num_services num_services_crit
                                                                                       num_services_ok num_services_pending num_services_unknown
                                                                                       num_services_warn worst_service_state/]);
        $servicegroupdata  = Thruk::Utils::array2hash($data, 'name');
    }

    return({
        'hosts'         => $hostdata,
        'services'      => $servicedata,
        'hostgroups'    => $hostgroupdata,
        'servicegroups' => $servicegroupdata
    });
}

##########################################################
sub _submit_results_to_core {
    my($self, $c, $results) = @_;
    return unless scalar @{$results} > 0;

    my $spool = $c->config->{'Thruk::Plugin::BP'}->{'spool_dir'};
    if($spool) {
        die("spool folder does not exist ".$spool.": ".$!) unless -d $spool;
        my $fh = File::Temp->new(
            TEMPLATE => "cXXXXXX",
            DIR      => $spool,
        );
        $fh->unlink_on_destroy(0);
        binmode($fh, ":encoding(UTF-8)");
        print $fh "### Active Check Result File ###\n";
        print $fh sprintf("file_time=%d\n\n",time);
        for my $r (@{$results}) {
            print $fh $r,"\n";
        }
        Thruk::Utils::IO::close($fh, $fh->filename);
        chmod(0664, $fh->filename); # make sure the core can read it

        my $file = $fh->filename.'.ok';
        sysopen my $t,$file,O_WRONLY|O_CREAT|O_NONBLOCK|O_NOCTTY
            or croak("Can't create $file : $!");
        Thruk::Utils::IO::close($t, $file);
    }

    return;
}

##########################################################

=head2 TO_JSON

    TO_JSON()

returns data needed to represent this module in json

=cut
sub TO_JSON {
    my($self) = @_;
    my $data = {};
    for my $key (@extra_json_keys, @stateful_keys, @saved_keys) {
        $data->{$key} = $self->{$key};
    }
    $data->{'nodes'} = [];
    for my $n (@{$self->{'nodes'}}) {
        push @{$data->{'nodes'}}, $n->TO_JSON();
    }
    return $data;
}

##########################################################


=head1 AUTHOR

Sven Nierlein, 2013, <sven.nierlein@consol.de>

=head1 LICENSE

This library is free software, you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

1;
